﻿#include "stdafx.h"

#ifndef __specstrings
#define __in
#define __in_z
#define __out_opt
#endif

class Module
{
    size_t m_ptr;
public:
    Module(HMODULE module) : m_ptr( reinterpret_cast<size_t>(module) )
    {
    }
    template<typename T> 
    const T* Ptr( size_t offset )
    {
        if(offset <= 0)
        {
            DebugBreak();
        }
        return reinterpret_cast<const T*>( m_ptr + offset );
    }
};

namespace APIHook 
{

HRESULT
Hook(__in HMODULE m,
        __in_z const char* dllName,
        __in_z const char* procName,
        __in PROC newProc,
        __out_opt PROC* originalProc)
{
    if( m == NULL )
    {
        return E_INVALIDARG;
    }
    if( dllName == NULL )
    {
        return E_INVALIDARG;
    }
    if( procName == NULL )
    {
        return E_INVALIDARG;
    }
    if( newProc == NULL )
    {
        return E_INVALIDARG;
    }

#ifndef _WIN64
    OSVERSIONINFO versionInfo;
    memset ( &versionInfo , NULL , sizeof ( OSVERSIONINFO ) ) ;
    versionInfo.dwOSVersionInfoSize = sizeof ( OSVERSIONINFO ) ;

    if( versionInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS &&
        reinterpret_cast<unsigned int>(m) >= 0x80000000 )
    {
        return E_FAIL;
    }
#endif

    const IMAGE_DOS_HEADER *dosHeader = reinterpret_cast<const IMAGE_DOS_HEADER *>( m );
    if( dosHeader->e_magic != IMAGE_DOS_SIGNATURE  )
    {
        return E_INVALIDARG;
    }

    Module module(m);

    const IMAGE_NT_HEADERS *ntHeader =    module.Ptr<IMAGE_NT_HEADERS>( dosHeader->e_lfanew );
    if( ntHeader->Signature != IMAGE_NT_SIGNATURE  )
    {
        return E_INVALIDARG;
    }

    const IMAGE_IMPORT_DESCRIPTOR *impDesc = 
      module.Ptr<IMAGE_IMPORT_DESCRIPTOR>(
        ntHeader->OptionalHeader.DataDirectory[ IMAGE_DIRECTORY_ENTRY_IMPORT ].VirtualAddress );

    for ( ; impDesc->Name != NULL; ++impDesc)
    {
        const char* moduleName = module.Ptr<const char>( impDesc->Name ) ;
        if ( _stricmp( moduleName , dllName ) == 0 )
        {
            break;
        }
    }

    if( impDesc->Name == NULL || 
        impDesc->FirstThunk == NULL || 
        impDesc->OriginalFirstThunk == NULL)
    {
        return S_FALSE;
    }

    for (  const IMAGE_THUNK_DATA
        *impNameTbl = module.Ptr<IMAGE_THUNK_DATA>( impDesc->OriginalFirstThunk  ),
        *impAddrTbl = module.Ptr<IMAGE_THUNK_DATA>( impDesc->FirstThunk  );
    impNameTbl->u1.Function != NULL;
    ++impNameTbl,  ++impAddrTbl )
    {
        if ( IMAGE_SNAP_BY_ORDINAL( impNameTbl->u1.Ordinal ) )
        {
            continue;
        }

        const IMAGE_IMPORT_BY_NAME *byName =
            module.Ptr<IMAGE_IMPORT_BY_NAME>( impNameTbl->u1.AddressOfData );

        const char* funcName = reinterpret_cast<const char*>(byName->Name);

        if( 0 != _stricmp( funcName, procName ) )
        {
            continue;
        }

        MEMORY_BASIC_INFORMATION memInfo;
        VirtualQuery( impAddrTbl,  &memInfo,  sizeof ( MEMORY_BASIC_INFORMATION ) );

        if ( 0 == VirtualProtect ( memInfo.BaseAddress,
            memInfo.RegionSize,
            PAGE_EXECUTE_READWRITE,
            &memInfo.Protect) )
        {
            return E_FAIL;
        }

        IMAGE_THUNK_DATA* thunk = const_cast<IMAGE_THUNK_DATA*>(impAddrTbl);

        PROC original = static_cast<PROC>( 
            InterlockedExchangePointer(
                reinterpret_cast<void**>( &(thunk->u1.Function) ), newProc ) );

        DWORD dummy;
        if ( 0 == VirtualProtect ( memInfo.BaseAddress,
            memInfo.RegionSize,
            memInfo.Protect,
            &dummy) )
        {
            return E_FAIL;
        }

        if( originalProc != NULL )
        {
            *originalProc = original;
        }

        return S_OK;
    }

    return S_FALSE;
}

} // end of nanespace "APIHook"

#ifndef __specstrings
#undef __in
#undef __in_z
#undef __out_opt
#endif